/*
 * spopup.c - simple popup DLL for mIRC
 * Copyright (c) 2006, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#define MAXHANDLES 256
#define MAXSTACK 16

typedef struct {
  DWORD  mVersion;
  HWND   mHwnd;
  BOOL   mKeep;
} LOADINFO;

static struct {
  HMENU menu;
  int id;
} stack[MAXSTACK];

static HMENU handles[MAXHANDLES];

static int hpos, spos;
static HWND hwnd;

static char *getnum(char *string, int *num) {
  int value = 0, filled = 0, neg = 0;
  char *ptr = string;

  if (ptr == NULL) return NULL;

  while (*ptr == ' ') ptr++;

  if (*ptr == '-') {
    neg = 1;
    ptr++;
  }

  for ( ; *ptr >= '0' && *ptr <= '9'; filled = 1)
   value = value * 10 + *ptr++ - '0';

  if (!filled) return NULL;

  if (*ptr == ' ') ptr++;
  else if (*ptr) return NULL;

  if (neg) value = -value;

  *num = value;
  return ptr;
}

static void pop(int depth) {
  while (spos > depth) {

    // gray out previous parent-items that contain no children, if any
    if (stack[spos].id > 0)
      EnableMenuItem(stack[spos - 1].menu,
       stack[spos].id, MF_BYCOMMAND | MF_GRAYED);

    spos--;
  }
}

static void destroy(void) {
  while (hpos >= 0) DestroyMenu(handles[hpos--]);
}

void __export __stdcall LoadDll(LOADINFO *li) {
  hpos = -1;
  hwnd = NULL;
}

int __export __stdcall UnloadDll(int timeout) {
  if (timeout == 1 && hwnd != NULL) return 0;

  destroy();

  return 1;
}

int __stdcall __export add(HWND mWnd, HWND aWnd, char *data, char *parms,
 BOOL show, BOOL nopause) {
  MENUITEMINFO mii;
  HMENU menu, submenu;
  char *ptr;
  int id, depth, style;

  // if we don't have a popup menu at all yet, initialize
  if (hpos < 0) {
    menu = CreatePopupMenu();
    handles[hpos = 0] = menu;
    stack[spos = 0].menu = menu;
    stack[spos].id = 0;
  }

  // parse the input
  ptr = getnum(data, &id);
  ptr = getnum(ptr, &depth);
  ptr = getnum(ptr, &style);
  if (hwnd != NULL || !ptr || !*ptr || id <= 0 || depth < 0) return 1;

  // initialize the menu item structure
  mii.cbSize = sizeof(mii);
  mii.fMask = MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_TYPE;
  mii.fType = (*ptr == '-' && !ptr[1]) ? MFT_SEPARATOR : MFT_STRING;
  mii.wID = id;
  mii.dwTypeData = ptr;
  mii.fState = 0;
  if (style & 1) mii.fState |= MFS_CHECKED;
  if (style & 2) mii.fState |= MFS_GRAYED;
  if (style & 4) mii.fState |= MFS_DEFAULT;
  if (style & 8) mii.fType |= MFT_RADIOCHECK;

  // decide at which depth to put this item at
  if (depth >= spos) depth = spos;
  if (depth < spos) pop(depth);

  // this is the parent item we're going to use
  menu = stack[spos].menu;
  stack[spos].id = 0;

  // if requested, create a new parent-item, initially with no children
  if (style & 16) {
    if (hpos == MAXHANDLES - 1 || spos == MAXSTACK - 1) return 1;

    submenu = CreatePopupMenu();
    mii.hSubMenu = submenu;
    handles[++hpos] = submenu;
    stack[++spos].menu = submenu;
    stack[spos].id = id;

    mii.fMask |= MIIM_SUBMENU;
  }

  // and insert the new item at the end of the parent menu
  InsertMenuItem(menu, GetMenuItemCount(menu), TRUE, &mii);

  return 1;
}

int __stdcall __export popup(HWND mWnd, HWND aWnd, char *data, char *parms,
 BOOL show, BOOL nopause) {
  int x, y, style, flags, res;
  HWND wnd;
  char *ptr;

  // make sure there's a menu to display, and it's not already opened
  if (hpos < 0 || hwnd != NULL) return 1;

  // gray out the last parent-items that contain no children, if any
  pop(0);

  // parse the input
  ptr = getnum(data, &x);
  ptr = getnum(ptr, &y);
  if (ptr == NULL) return 1;

  ptr = getnum(ptr, &style);
  if (ptr == NULL) style = 0;

  if (getnum(ptr, (int *)&wnd) == NULL) wnd = aWnd;

  flags = TPM_NONOTIFY | TPM_RETURNCMD;
  if (style & 1) flags |= TPM_RIGHTBUTTON;
  if (style & 2) flags |= TPM_CENTERALIGN;
  if (style & 4) flags |= TPM_RIGHTALIGN;
  if (style & 8) flags |= TPM_VCENTERALIGN;
  if (style & 16) flags |= TPM_BOTTOMALIGN;

  // create the popup menu, and wait for it to finish
  hwnd = wnd;
  res = TrackPopupMenuEx(stack[0].menu, flags, x, y, hwnd, NULL);
  hwnd = NULL;

  // destroy all the parts of the popup menu
  destroy();

  // and return the selected ID, or 0 if none
  wsprintf(data, "%d", res);

  return 3;
}

int __stdcall __export clear(HWND mWnd, HWND aWnd, char *data, char *parms,
 BOOL show, BOOL nopause) {
  // if no popup is active, remove all popup items added so far
  if (hwnd == NULL) destroy();

  return 1;
}

int __stdcall __export cancel(HWND mWnd, HWND aWnd, char *data, char *parms,
 BOOL show, BOOL nopause) {
  // sending WM_CANCELMODE to the window that owns the popup, will cancel it
  if (hwnd != NULL) PostMessage(hwnd, WM_CANCELMODE, 0, 0);

  return 1;
}
